# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Takes in OpenQASM 3 code and outputs an ASCII circuit representation

"""

import re

import numpy as np

controlled_gates = [
    "cx",
    "c3x",
    "c3sx",
    "c4x",
    "ccx",
    "ch",
    "crx",
    "cry",
    "crz",
    "cs",
    "csdg",
    "cswap",
    "csx",
    "cu",
    "cu1",
    "cu3",
    "cx",
    "cy",
    "cz",
    "ccz",
]

multi_qubit_gates = [
    "cp",
    "dcx",
    "ecr",
    "rccx",
    "rc3x",
    "rxx",
    "ryy",
    "rzz",
    "rzx",
    "xx_minus_yy",
    "xx_plus_yy",
    "iswap",
    "swap",
]

single_qubit_gates = [
    "h",
    "id",
    "p",
    "r",
    "rx",
    "ry",
    "rz",
    "s",
    "sdg",
    "sx",
    "sxdg",
    "t",
    "tdg",
    "U",
    "u1",
    "u2",
    "u3",
    "x",
    "y",
    "z",
]

q_c_reg_gates = ["m"]

all_gates = single_qubit_gates + multi_qubit_gates + controlled_gates + q_c_reg_gates


def is_valid_gate(line):
    """Checks if input string line is a valid OPENQASM gate"""
    if len(line) < 2:
        return False
    if len(line) > 2 and line[:2] == "//":
        return False
    if len(line) > 7 and line[:7] == "include":
        return False
    if len(line) > 8 and line[:8] == "OPENQASM":
        return False
    return True


def parse_gate_type(line, gates):
    """Returns gate type and params of a line"""
    params = None
    if "measure" in line:
        return ("m", params)

    gate_type = line.split(" ")[0]

    if "(" in line:
        gate_type = line.split("(")[0]
        params = re.search(r"\(.*\)", line)[0][1:-1]

    if gate_type not in gates:
        return (None, None)
    return (gate_type, params)


class Gate:
    """An individual gate in the circuit"""

    def __init__(self, line, num_qregs, num_cregs):
        """Initializes a gate"""

        assert is_valid_gate(line), "Invalid gate: " + str(line)

        self.gate_type, self.params = parse_gate_type(line, all_gates)

        assert self.gate_type is not None, "Gate is not found in list of possible gates"

        self.num_qregs = num_qregs
        self.num_cregs = num_cregs
        self.qregs = self.parse_qregs(line)
        self.cregs = self.parse_cregs(line)

    def parse_qregs(self, line):
        """Gets the quantum registers associated with the gate"""
        matches = re.findall(r"q\[.*?\]", line)
        qregs = []
        for match in matches:
            qregs.append(int(match[2:-1]))
        return qregs

    def parse_cregs(self, line):
        """Gets the classical registers associated with the gate"""
        matches = re.findall(r"b\[.*?\]", line)
        cregs = []
        for match in matches:
            cregs.append(int(match[2:-1]))
        return cregs

    def get_height(self):
        """Gets the height of the gate"""
        if self.gate_type in single_qubit_gates:
            return 1
        if self.gate_type in q_c_reg_gates:
            return self.num_qregs - min(self.qregs) + max(self.cregs) + 1
        return max(self.qregs) - min(self.qregs) + 1

    def get_width(self):
        """Gets the width of the gate"""
        gate_str = self.gate_type

        if self.params is not None:
            gate_str += "(" + self.params + ")"

        if gate_str in ("swap", "cswap"):
            return 3

        return len(gate_str) + 4

    def apply_gate_at_pos(self, gate_str, mat, pos):
        """Adds the gate at the given y coordinate in the matrix representation"""
        mat[pos, :] = ["|"] + ["-"] * (len(gate_str) + 2) + ["|"]
        mat[pos + 1, :] = ["|"] + [" "] + list(gate_str) + [" "] + ["|"]
        mat[pos + 2, :] = ["|"] + ["-"] * (len(gate_str) + 2) + ["|"]

        return mat

    def mat(self):
        """Outputs a matrix representation of the gate"""
        height = self.get_height()
        gate_str = self.gate_type
        mat = np.array([])

        if self.params is not None:
            gate_str += "(" + self.params + ")"

        if height == 1:
            mat = np.empty((3 * height, len(gate_str) + 4), dtype="str")
            mat = self.apply_gate_at_pos(gate_str, mat, 0)

        elif self.gate_type == "m":
            mat = np.empty((3 * height, len(gate_str) + 4), dtype="str")
            mat = self.apply_gate_at_pos(gate_str, mat, 0)

            c_reg_line = (
                [" "] * int((len(gate_str) + 4) / 2) + ["║"] + [" "] * int((len(gate_str) + 4) / 2)
            )
            mat[3 : 3 * height - 1, :] = [c_reg_line] * (3 * height - 4)
            mat[3 * height - 1, :] = [" "] * (len(gate_str) + 4)

        elif self.gate_type in controlled_gates:
            if gate_str == "cswap":
                mat = np.full((3 * height, 3), " ", dtype="str")
                mat[1:-1, 1] = ["|"] * (mat.shape[1] - 2)
                mat[(self.qregs[0] - min(self.qregs)) * 3 + 1, :2] = ["-", "■"]
                mat[(self.qregs[1] - min(self.qregs)) * 3 + 1, :2] = ["-", "X"]
                mat[(self.qregs[2] - min(self.qregs)) * 3 + 1, :2] = ["-", "X"]
                return mat

            mat = np.full((3 * height, len(gate_str) + 4), " ", dtype="str")
            mid = int(mat.shape[1] / 2)
            mat[1:-1, mid] = ["|"] * (mat.shape[0] - 2)

            for qreg in range(len(self.qregs) - 1):
                mat[(self.qregs[qreg] - min(self.qregs)) * 3 + 1, : mid + 1] = ["-"] * mid + ["■"]

            for qreg in range(height):
                if qreg != self.qregs[-1]:
                    mat[3 * qreg + 1, :mid] = ["-"] * mid
            mat = self.apply_gate_at_pos(gate_str, mat, (self.qregs[-1] - min(self.qregs)) * 3)

        elif self.gate_type in multi_qubit_gates:
            if gate_str == "swap":
                mat = np.empty((3 * height, 3), dtype="str")
                mat[0, :] = [" "] * 3
                mat[1, :] = ["-", "X", " "]
                mat[2 : 3 * height - 2, :] = [[" ", "|", " "]] * (3 * height - 4)
                mat[3 * height - 2, :] = ["-", "X", " "]
                mat[3 * height - 1, :] = [" "] * 3
                mat[range(1, 3 * height, 3), 0] = ["-"] * height

            else:
                mat = np.full((3 * height, len(gate_str) + 4), " ", dtype="str")
                mat[0, :] = ["|"] + ["-"] * (mat.shape[1] - 2) + ["|"]
                mat[1:-1, :] = ["|"] + [" "] * (mat.shape[1] - 2) + ["|"]
                for qreg_idx, _ in enumerate(self.qregs):
                    mat[(self.qregs[qreg_idx] - min(self.qregs)) * 3 + 1, 1] = qreg_idx
                mat[int(mat.shape[0] / 2), mat.shape[1] - len(gate_str) - 1 : -1] = list(gate_str)
                mat[-1, :] = ["|"] + ["-"] * (mat.shape[1] - 2) + ["|"]

        return mat


def get_collision(circuit, g):
    """Gets the collision position of a gate"""
    pos = circuit.shape[1] - 1
    collision = False
    reg_idxes = range(3 * min(g.qregs), 3 * (g.get_height() + min(g.qregs)))
    while not collision and pos > 0:
        for reg_idx in reg_idxes:
            if circuit[reg_idx][pos] != " ":
                collision = True
                break
        if collision:
            break
        pos -= 1

    if pos != 0:
        pos += 3

    return pos, reg_idxes


def add_gate(circuit, g, pos, reg_idxes):
    """Adds a gate to the circuit"""

    while pos + g.get_width() >= circuit.shape[1]:
        circuit = np.concatenate(
            [circuit, np.full((circuit.shape[0], circuit.shape[1]), " ", dtype=str)], axis=1
        )

    circuit[reg_idxes, pos : pos + g.get_width()] = g.mat()

    return circuit


def add_moment(circuit, moment):
    """Adds a moment to the circuit"""
    collisions = [get_collision(circuit, g)[0] for g in moment]
    pos = max(collisions)
    widths = [g.get_width() for g in moment]
    max_w = max(widths)

    for g in moment:
        _, reg_idxes = get_collision(circuit, g)
        centered_moment_pos = pos + int((max_w - g.get_width()) / 2)
        circuit = add_gate(circuit, g, centered_moment_pos, reg_idxes)
        circuit = add_wires(circuit, g, g.num_qregs, centered_moment_pos)

    return circuit


def get_circuit_height(qasm_str):
    """Gets the number of qubit and classical registers in the circuit"""
    num_qregs = 0
    num_cregs = 0
    for line in qasm_str.split("\n"):
        qreg_match = re.match(r"qreg (\w+)\[\d+\];", line)
        if qreg_match:
            num_qregs = int(re.search(r"\[(\d+)\]", line).group(1))
            continue
        creg_match = re.match(r"creg (\w+)\[\d+\];", line)
        if creg_match:
            num_cregs = int(re.search(r"\[(\d+)\]", line).group(1))
            continue
        if re.match(r"qubit\[.*\]", line):
            num_qregs = int(re.match(r"qubit\[.*\]", line)[0][6:-1])
            continue
        if re.match(r"bit\[.*\]", line):
            num_cregs = int(re.match(r"bit\[.*\]", line)[0][4:-1])
            continue

    return num_qregs, num_cregs


def get_collision_before_pos(circuit, x, y):
    """Gets the position of nearest collision with another gate"""
    while circuit[y, x] == " " and x >= 0:
        x -= 1
    return x


def add_wires(circuit, g, num_qregs, pos):
    """Adds wires to connect gates"""
    if pos == 0:
        return circuit

    start_wire = min(g.qregs)
    end_wire = max(g.qregs)
    if g.gate_type == "m":
        end_wire = num_qregs
    for y in range(3 * start_wire + 1, 3 * end_wire + 4, 3):
        while circuit[y, pos] == " ":
            pos += 1
        collision_pos = get_collision_before_pos(circuit, pos - 1, y)
        circuit[y, collision_pos + 1 : pos] = ["-"] * (pos - collision_pos - 1)

    return circuit


def add_cregs(circuit, num_cregs, end_pos):
    """Adds classical registers to the circuit"""
    for bit in range(num_cregs):
        y = -2 - (bit * 3)
        circuit[y, :end_pos] = ["="] * end_pos

    return circuit


def extend_qregs(circuit, num_qregs, end_pos):
    """Extends the quantum registers to the end of the circuit"""
    for qreg in range(num_qregs):
        eol_regex = re.search(r"(\S)\s*$", "".join(circuit[3 * qreg + 1]))
        eol_pos = eol_regex.regs[1][1]
        circuit[3 * qreg + 1, eol_pos:end_pos] = ["-"] * (end_pos - eol_pos)

    return circuit


def can_add_gate(m_qregs, m_cregs, gate):
    """Checks if a gate can be added to a moment"""
    if len(gate.qregs) != 0:
        min_qreg = min(gate.qregs)
        max_qreg = max(gate.qregs) if gate.gate_type not in q_c_reg_gates else gate.num_qregs
        for qreg in range(min_qreg, max_qreg + 1):
            if qreg in m_qregs:
                return False

    if len(gate.cregs) != 0:
        min_creg = min(gate.cregs) if gate.gate_type not in q_c_reg_gates else 0
        max_creg = max(gate.cregs)
        for creg in range(min_creg, max_creg + 1):
            if creg in m_cregs:
                return False

    return True


def get_moments(gates):
    """Splits up the circuit into moments"""
    moments = []
    while gates != []:
        gate = gates.pop(0)
        moment = [gate]
        m_qregs = list(gate.qregs)
        m_cregs = list(gate.cregs)
        gate_idx = 0
        while gate_idx < len(gates):
            gate = gates[gate_idx]
            if can_add_gate(m_qregs, m_cregs, gate):
                moment.append(gates.pop(gate_idx))
                gate_idx -= 1

            for qreg in gate.qregs:
                m_qregs.append(qreg)
            for creg in gate.cregs:
                m_cregs.append(creg)

            gate_idx += 1
        moments.append(moment)
    return moments


def _qasm3_drawer(qasm_str: str) -> str:
    """Iterates over the input string and returns the gates"""
    num_qregs, num_cregs = get_circuit_height(qasm_str)
    height = (num_qregs + num_cregs) * 3
    circuit = np.full((height, 50), " ", dtype=str)

    gates = []
    for line in qasm_str.split("\n"):
        valid = is_valid_gate(line)
        gate_type, _ = parse_gate_type(line, all_gates)
        if valid and gate_type is not None:
            g = Gate(line, num_qregs, num_cregs)
            gates.append(g)

    moments = get_moments(gates)
    for moment in moments:
        circuit = add_moment(circuit, moment)

    #    circuit, pos = add_gate(circuit, g)
    #    circuit = add_wires(circuit, g, num_qregs, pos)

    end_pos = circuit.shape[1]
    while np.all(circuit[:, end_pos - 1] == [" "] * circuit.shape[0]):
        end_pos -= 1
    end_pos += 1

    circuit = add_cregs(circuit, num_cregs, end_pos)
    circuit = extend_qregs(circuit, num_qregs, end_pos)

    padding = 5

    out = ""
    reg_idx = 0
    is_q_reg = True
    for row_idx, _ in enumerate(circuit):
        line = ""
        if row_idx % 3 == 1:
            if reg_idx < num_qregs and is_q_reg:
                line += ("q" + str(reg_idx)).ljust(padding, "-")
            elif is_q_reg:
                reg_idx = 0
                is_q_reg = False

            if not is_q_reg:
                line += ("c" + str(reg_idx)).ljust(padding, "=")
            reg_idx += 1
        else:
            line += " " * padding
        line += "".join(circuit[row_idx]).rstrip() + "\n"
        if line.strip() != "":
            out += line

    return out[:-1]


def qasm3_drawer(qasm_str: str) -> None:
    """Draws the circuit from the input string"""
    print(_qasm3_drawer(qasm_str))



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for converting Braket circuits to OpenQASM 3.0

"""
from braket.circuits import Circuit
from braket.circuits.serialization import IRType

from qbraid.passes.qasm import remove_measurements
from qbraid.transpiler.annotations import weight


@weight(1)
def braket_to_qasm3(circuit: Circuit) -> str:
    """Converts a ``braket.circuits.Circuit`` to an OpenQASM 3.0 string.

    Args:
        circuit: Amazon Braket quantum circuit

    Returns:
        The OpenQASM 3.0 string equivalent to the circuit

    Raises:
        ProgramConversionError: If braket to qasm conversion fails

    """
    has_measurements = any(str(instr.operator) == "Measure" for instr in circuit.instructions)
    qasm_program = circuit.to_ir(IRType.OPENQASM).source
    if not has_measurements:
        qasm_program = remove_measurements(qasm_program)
    return qasm_program



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing sub-modules for interfacing with
various quantum software libraries and program types.

"""
import importlib
import sys
import warnings

from qbraid.programs.gate_model._model import GateModelProgram as GateModelProgram

warnings.warn(
    "The 'qbraid.programs.circuits' module has been renamed to "
    "'qbraid.programs.gate_model'. Please update your imports. "
    "This backward compatibility layer will be removed in a future release.",
    DeprecationWarning,
    stacklevel=2,
)

# Redirect imports from 'circuits' to 'gate_model'
module_name = sys.modules[__name__]

submodules = [
    "_model",
    "braket",
    "cirq",
    "ionq",
    "pennylane",
    "pyquil",
    "pytket",
    "qasm2",
    "qasm3",
    "qiskit",
]

# Dynamically import and register submodules
for submodule in submodules:
    new_module_name = f"qbraid.programs.gate_model.{submodule}"
    sys.modules[f"{__name__}.{submodule}"] = importlib.import_module(new_module_name)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for providing transforamtions to ensure OpenQASM 3 compatibility
across various other quantum software frameworks.

"""
import math
import re
from functools import reduce
from typing import Union

from openqasm3 import dumps, parse
from openqasm3.ast import Include, Program, QuantumGate, QuantumMeasurementStatement, Statement
from openqasm3.parser import QASM3ParsingError

from qbraid._logging import logger

GATE_DEFINITIONS = {
    "iswap": """
gate iswap _gate_q_0, _gate_q_1 {
  s _gate_q_0;
  s _gate_q_1;
  h _gate_q_0;
  cx _gate_q_0, _gate_q_1;
  cx _gate_q_1, _gate_q_0;
  h _gate_q_1;
}""",
    "sxdg": """
gate sxdg _gate_q_0 {
  s _gate_q_0;
  h _gate_q_0;
  s _gate_q_0;
}""",
}


def insert_gate_def(qasm3_str: str, gate_name: str, force_insert: bool = False) -> str:
    """Add gate definitions to an Open0QASM 3 string.

    Args:
        qasm3_str (str): QASM 3.0 string.
        gate_name (str): Name of the gate to insert.
        force_insert (bool): If True, the gate definition will be added
            even if the gate is never referenced. Defaults to False.

    Returns:
        str: QASM 3.0 string with gate definition.

    Raises:
        ValueError: If the gate definition is not found.
    """
    defn = GATE_DEFINITIONS.get(gate_name)

    if defn is None:
        raise ValueError(
            f"Gate {gate_name} definition not found. "
            f"Available gate definitions include: {set(GATE_DEFINITIONS.keys())}"
        )

    if not force_insert and gate_name not in qasm3_str:
        return qasm3_str

    lines = qasm3_str.splitlines()

    insert_index = 0
    for i, line in enumerate(lines):
        if "include" in line or "OPENQASM" in line:
            insert_index = i + 1
            break

    lines.insert(insert_index, defn.strip())

    return "\n".join(lines)


def _normalize_case_insensitive_map(gate_mappings: dict[str, str]) -> dict[str, str]:
    """Normalize gate_mappings keys to lowercase and check for duplicates."""
    lowercased_map = {}
    for key, value in gate_mappings.items():
        lower_key = key.lower()
        if lower_key in lowercased_map:
            raise ValueError(
                f"Duplicate key detected after lowercasing: '{lower_key}'. Use case_sensitive=True."
            )
        lowercased_map[lower_key] = value
    return lowercased_map


def _replace_gate_in_statement(
    statement: Statement, gate_mappings: dict[str, str], case_sensitive: bool
) -> QuantumGate:
    """Replace gate names in a single statement if applicable."""
    if isinstance(statement, QuantumGate):
        gate_name = statement.name.name
        lookup_name = gate_name if case_sensitive else gate_name.lower()
        if lookup_name in gate_mappings:
            statement.name.name = gate_mappings[lookup_name]
    return statement


def _replace_gate_names(
    program: Program, gate_mappings: dict[str, str], case_sensitive: bool
) -> Program:
    """Replace occurrences of specified gate names in a openqasm3.ast.Program."""
    if not case_sensitive:
        gate_mappings = _normalize_case_insensitive_map(gate_mappings)

    statements = [
        _replace_gate_in_statement(stmnt, gate_mappings, case_sensitive)
        for stmnt in program.statements
    ]

    return Program(statements=statements, version=program.version)


def replace_gate_names(
    qasm: str, gate_mappings: dict[str, str], case_sensitive: bool = False
) -> str:
    """Replace occurrences of specified gate names in a QASM program string.

    Args:
        qasm (str): The QASM program as a string.
        gate_mappings (dict[str, str]): A dictionary mapping old gate names (keys)
            to new gate names (values).
        case_sensitive (bool): Whether the gate name replacement should be case-sensitive.
            Defaults to False.

    Returns:
        str: The modified QASM program with the gate names replaced.

    Raises:
        ValueError: If duplicate keys are found in the gate_mappings after lowercasing.
    """
    program = parse(qasm)

    program_out = _replace_gate_names(program, gate_mappings, case_sensitive)
    version_major = program_out.version.split(".")[0]
    qasm_out = dumps(program_out)

    if int(version_major) == 2:
        qasm_out = declarations_to_qasm2(qasm_out)

    return qasm_out


def add_stdgates_include(qasm_str: str) -> str:
    """Add 'include "stdgates.inc";' to the QASM string if it is missing."""
    if 'include "stdgates.inc";' in qasm_str:
        return qasm_str

    lines = qasm_str.splitlines()

    for i, line in enumerate(lines):
        if "OPENQASM" in line:
            lines.insert(i + 1, 'include "stdgates.inc";')
            break

    return "\n".join(lines)


def remove_stdgates_include(qasm: str) -> str:
    """Remove 'include "stdgates.inc";' from the QASM string."""
    return qasm.replace('include "stdgates.inc";', "")


def _evaluate_expression(match):
    """Helper function for simplifying arithmetic expressions within parentheses."""
    expr = match.group(1)
    try:
        simplified_value = eval(expr)  # pylint: disable=eval-used
        return f"({simplified_value})"
    except SyntaxError:
        return match.group(0)


def simplify_arithmetic_expressions(qasm_str: str) -> str:
    """Simplifies arithmetic expressions within parentheses in a QASM string."""

    pattern = r"\(([0-9+\-*/. ]+)\)"

    return re.sub(pattern, _evaluate_expression, qasm_str)


def convert_qasm_pi_to_decimal(qasm: str) -> str:
    """Convert all instances of 'pi' in the QASM string to their decimal value."""

    pattern = r"(?<![a-zA-Z])(\d*\.?\d*\s*[*/+-]\s*)?pi(\s*[*/+-]\s*\d*\.?\d*)?(?![a-zA-Z])"

    gate_defs = set()

    try:
        program = parse(qasm)

        for statement in program.statements:
            if isinstance(statement, QuantumGate):
                name = statement.name.name
                if "pi" in name:
                    gate_defs.add(name)
    except QASM3ParsingError as err:
        logger.debug("Failed to parse QASM program for pi conversion: %s", err)

    def replace_with_decimal(match: re.Match) -> str:
        expr: str = match.group()
        start = match.start()
        end = match.end()

        for gate_def in gate_defs:
            if gate_def in qasm[max(0, start - len(gate_def)) : end]:
                return expr  # pragma: no cover

        expr_with_pi_as_decimal = expr.replace("pi", str(math.pi))
        try:
            value = eval(expr_with_pi_as_decimal)  # pylint: disable=eval-used
        except SyntaxError:
            return expr
        return str(value)

    return re.sub(pattern, replace_with_decimal, qasm)


def has_redundant_parentheses(qasm_str: str) -> bool:
    """Checks if a QASM string contains gate parameters with redundant parentheses."""

    pattern = r"\w+\(\(\s*[-+]?\d+(\.\d*)?\s*\)\)"
    if re.search(pattern, qasm_str):
        return True

    pattern_neg = r"\w+\(-\(\d*\.?\d+\)\)"
    if re.search(pattern_neg, qasm_str):
        return True

    return False


def remove_spaces_in_parentheses(expression: str) -> str:
    """Removes all spaces inside parentheses in an expression."""
    parenthesized_parts = re.findall(r"\(.*?\)", expression)

    for part in parenthesized_parts:
        cleaned_part = part.replace(" ", "")
        expression = expression.replace(part, cleaned_part)

    return expression


def simplify_parentheses_in_qasm(qasm_str: str) -> str:
    """Simplifies unnecessary parentheses around numbers in QASM strings."""

    lines = qasm_str.splitlines()
    simplified_lines = []

    pattern = r"\(\s*([-+]?\s*\d+(\.\d*)?)\s*\)"

    def simplify(match):
        return match.group(1).replace(" ", "")

    for line in lines:
        if has_redundant_parentheses(line):
            line = re.sub(pattern, simplify, line)
        simplified_lines.append(line)

    return "\n".join(simplified_lines)


def compose(*functions):
    """Compose multiple functions left to right."""

    def compose_two(f, g):
        return lambda x: g(f(x))

    return reduce(compose_two, functions, lambda x: x)


def normalize_qasm_gate_params(qasm: str) -> str:
    """Normalize the parameters of the gates in the QASM string using function composition."""
    transform_qasm = compose(
        convert_qasm_pi_to_decimal, simplify_arithmetic_expressions, simplify_parentheses_in_qasm
    )
    return transform_qasm(qasm)


def declarations_to_qasm2(qasm: str) -> str:
    """Converts QASM 3.0 qubit and bit declarations to QASM 2.0 qreg and creg declarations."""
    for declaration_type, replacement_type in [("qubit", "qreg"), ("bit", "creg")]:
        pattern = rf"{declaration_type}\[(\d+)\]\s+(\w+);"
        replacement = rf"{replacement_type} \2[\1];"
        qasm = re.sub(pattern, replacement, qasm)

    return qasm


def remove_qasm_barriers(qasm_str: str) -> str:
    """Returns a copy of the input QASM with all barriers removed.

    Args:
        qasm_str: QASM to remove barriers from.
    """
    quoted_re = r"(?:\"[^\"]*?\")"
    statement_re = r"((?:[^;{}\"]*?" + quoted_re + r"?)*[;{}])?"
    comment_re = r"(\n?//[^\n]*(?:\n|$))?"
    statements_comments = re.findall(statement_re + comment_re, qasm_str)

    lines = []
    for statement, comment in statements_comments:
        if re.match(r"^\s*barrier(?:(?:\s+)|(?:;))", statement) is None:
            lines.append(statement + comment)
    return "".join(lines)


def remove_measurements(program: Union[Program, str]) -> str:
    """Remove all measurement operations from the program."""
    program = parse(program) if isinstance(program, str) else program
    statements = [
        statement
        for statement in program.statements
        if not isinstance(statement, QuantumMeasurementStatement)
    ]
    program_out = Program(statements=statements, version=program.version)
    program_str = dumps(program_out)
    if float(program.version) == 2.0:
        program_str = declarations_to_qasm2(program_str)
    return program_str


def remove_include_statements(program: Union[Program, str]) -> str:
    """Remove all include statements from the program."""
    program = parse(program) if isinstance(program, str) else program
    statements = [
        statement for statement in program.statements if not isinstance(statement, Include)
    ]
    program_out = Program(statements=statements, version=program.version)
    program_str = dumps(program_out)
    if float(program.version) == 2.0:
        program_str = declarations_to_qasm2(program_str)
    return program_str



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module submitting and managing jobs through IBM and IBM backends.

"""
import warnings

from .ibm import QiskitBackend, QiskitJob, QiskitRuntimeProvider

warnings.warn(
    "The 'qbraid.runtime.qiskit' module is deprecated. Please use 'qbraid.runtime.ibm' instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = ["QiskitBackend", "QiskitJob", "QiskitRuntimeProvider"]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Azure Provider class for retrieving all Azure backends.

"""
import warnings
from typing import Optional

from azure.identity import ClientSecretCredential
from azure.quantum import Workspace
from azure.quantum.target import Target
from qbraid_core._import import LazyLoader

from qbraid._caching import cached_method
from qbraid.programs import ExperimentType, IonQDict, ProgramSpec
from qbraid.runtime.exceptions import ResourceNotFoundError
from qbraid.runtime.profile import TargetProfile
from qbraid.runtime.provider import QuantumProvider

from .device import AzureQuantumDevice
from .io_format import InputDataFormat

pyquil = LazyLoader("pyquil", globals(), "pyquil")
pyqir = LazyLoader("pyqir", globals(), "pyqir")


class AzureQuantumProvider(QuantumProvider):
    """
    Manages interactions with Azure Quantum services, encapsulating API calls,
    handling Azure Storage, and managing sessions.

    Attributes:
        workspace (Workspace): The configured Azure Quantum workspace.
    """

    def __init__(self, workspace: Workspace, credential: Optional[ClientSecretCredential] = None):
        """
        Initializes an AzureQuantumProvider instance with a specified Workspace. It sets the
        credential for the workspace if it is not already set and a credential is provided.

        Args:
            workspace (Workspace): An initialized Azure Quantum Workspace object.
            credential (Optional[ClientSecretCredential]): Optional credential to be used
                if the workspace lacks one.
        """
        if not workspace.credential:
            if credential:
                workspace.credential = credential
            else:
                warnings.warn(
                    "No credential provided; interactive authentication via a "
                    "web browser may be required. To avoid interactive authentication, "
                    "provide a ClientSecretCredential."
                )
        workspace.append_user_agent("qbraid")
        self._workspace = workspace

    @property
    def workspace(self) -> Workspace:
        """Get the Azure Quantum workspace."""
        return self._workspace

    def _build_profile(self, target: Target) -> TargetProfile:
        """Builds a profile for an Azure device.

        Args:
            target (Target): The Target object to build profile from.

        Returns:
            TargetProfile: The completed profile.
        """
        device_id = target.name
        provider_name = target.provider_id
        simulator = device_id.split(".", 1)[1] != "qpu"
        capability = target.capability
        input_data_format = target.input_data_format
        output_data_format = target.output_data_format
        content_type = target.content_type

        if input_data_format == InputDataFormat.MICROSOFT.value:
            program_spec = ProgramSpec(
                pyqir.Module, alias="pyqir", to_ir=lambda module: module.bitcode
            )
        elif input_data_format == InputDataFormat.IONQ.value:
            program_spec = ProgramSpec(IonQDict, alias="ionq", to_ir=lambda ionq_dict: ionq_dict)
        elif input_data_format == InputDataFormat.QUANTINUUM.value:
            program_spec = ProgramSpec(str, alias="qasm2", to_ir=lambda qasm: qasm)
        elif input_data_format == InputDataFormat.RIGETTI.value:
            program_spec = ProgramSpec(
                pyquil.Program, alias="pyquil", to_ir=lambda program: program.out()
            )
        else:
            program_spec = None
            warnings.warn(f"Unrecognized input data format: {input_data_format}")

        return TargetProfile(
            device_id=device_id,
            simulator=simulator,
            provider_name=provider_name,
            capability=capability,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            content_type=content_type,
            experiment_type=ExperimentType.GATE_MODEL,
            program_spec=program_spec,
        )

    @cached_method
    def get_devices(self, **kwargs) -> list[AzureQuantumDevice]:
        """Get all Azure Quantum devices.

        Args:
            **kwargs: Filters for the devices to retrieve.

        Returns:
            list[AzureQuantumDevice]: The Azure Quantum devices.

        """
        targets = self.workspace.get_targets(**kwargs)
        if isinstance(targets, Target):
            targets = [targets]

        if len(targets) == 0:
            if len(kwargs) > 0:
                raise ValueError("No devices found with the specified filters.")
            raise ResourceNotFoundError("No devices found.")

        return [
            AzureQuantumDevice(self._build_profile(target), self.workspace) for target in targets
        ]

    @cached_method
    def get_device(self, device_id: str) -> AzureQuantumDevice:
        """Get a specific Azure Quantum device.

        Args:
            device_id (str): The ID of the device to retrieve.

        Returns:
            AzureQuantumDevice: The Azure Quantum device.

        """
        target = self.workspace.get_targets(name=device_id)
        if not target:
            raise ValueError(f"Device {device_id} not found.")
        return AzureQuantumDevice(self._build_profile(target), self.workspace)

    def __hash__(self):
        if not hasattr(self, "_hash"):
            object.__setattr__(
                self, "_hash", hash((self._workspace.credential, self._workspace.user_agent))
            )
        return self._hash  # pylint: disable=no-member



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM 3 conversion extras.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import requires_extras

qbraid_qir = LazyLoader("qbraid_qir", globals(), "qbraid_qir")

if TYPE_CHECKING:
    import pyqir

    from qbraid.programs.typer import Qasm3StringType


@requires_extras("qbraid_qir")
def qasm3_to_pyqir(program: Qasm3StringType) -> pyqir.Module:
    """Returns a PyQIR module equivalent to the input OpenQASM 3 program.

    Args:
        program (str): OpenQASM 3 program circuit to convert to PyQIR module.

    Returns:
        pyqir.Module: module equivalent to input OpenQASM 3 program.
    """
    return qbraid_qir.qasm3.qasm3_to_qir(program)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module used for loading entry points and other dynamic imports.

"""
import importlib.metadata
import sys
from typing import Optional, Type

try:
    import pkg_resources
except ImportError:  # pragma: no cover
    pkg_resources = None

from .exceptions import QbraidError


def get_entrypoints(module: str) -> dict[str, object]:
    """
    Retrieves entry points for a given module.

    Args:
        module (str): The name of the module to retrieve entry points for.

    Returns:
        dict[str, object]: Dictionary of entry point names and their associated entry point objects.
    """
    group = f"qbraid.{module}"

    if sys.version_info >= (3, 10):
        entry_points = importlib.metadata.entry_points().select(group=group)
    else:
        entry_points = pkg_resources.iter_entry_points(group=group)

    return {ep.name: ep for ep in entry_points}


def load_entrypoint(module: str, name: str) -> Optional[Type]:
    """
    Load an entrypoint given its module and name.

    Args:
        module (str): Module of entrypoint to load, e.g., "programs"
        name (str): Name of the entrypoint to load within the module.

    Returns:
        Optional[Type]: The loaded entry point class, or None if not found or failed to load.

    Raises:
        ValueError: If the specified entry point cannot be found.
        QbraidError: If the specified entry point fails to load.
    """
    try:
        entry_points = get_entrypoints(module)
        entry_point = entry_points[name]
        return entry_point.load()
    except KeyError as err:
        raise ValueError(f"Entrypoint '{name}' not found in module '{module}'.") from err
    except Exception as err:
        raise QbraidError(f"Failed to load entrypoint '{name}' from module '{module}'.") from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Mdule submiting and managing quantm tasks through AWS
and Amazon Braket supported devices.

"""
import warnings

from .aws import BraketDevice, BraketProvider, BraketQuantumTask

warnings.warn(
    "The 'qbraid.runtime.braket' module is deprecated. Please use 'qbraid.runtime.aws' instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = ["BraketProvider", "BraketDevice", "BraketQuantumTask"]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining BraketQuantumTask Class

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from braket.aws import AwsQuantumTask
from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import (
    AnalogHamiltonianSimulationQuantumTaskResult,
)
from braket.tasks.gate_model_quantum_task_result import GateModelQuantumTaskResult

from qbraid._logging import logger
from qbraid.runtime.enums import JobStatus
from qbraid.runtime.exceptions import JobStateError
from qbraid.runtime.job import QuantumJob
from qbraid.runtime.result import Result
from qbraid.runtime.result_data import AhsResultData, GateModelResultData

from .result_builder import BraketAhsResultBuilder, BraketGateModelResultBuilder
from .tracker import get_quantum_task_cost

if TYPE_CHECKING:
    from decimal import Decimal

    from braket.aws import AwsSession

AWS_TASK_STATUS_MAP = {
    "CREATED": JobStatus.INITIALIZING,
    "QUEUED": JobStatus.QUEUED,
    "RUNNING": JobStatus.RUNNING,
    "CANCELLING": JobStatus.CANCELLING,
    "CANCELLED": JobStatus.CANCELLED,
    "COMPLETED": JobStatus.COMPLETED,
    "FAILED": JobStatus.FAILED,
}


class AmazonBraketVersionError(Exception):
    """Exception raised for Amazon Braket SDK errors due to versioning."""


class BraketQuantumTask(QuantumJob):
    """Wrapper class for Amazon Braket ``QuantumTask`` objects."""

    def __init__(self, task_id: str, task: Optional[AwsQuantumTask] = None, **kwargs):
        """Create a BraketQuantumTask."""

        super().__init__(task_id, **kwargs)
        self._task = task or AwsQuantumTask(task_id)

    def status(self):
        """Returns status from Braket QuantumTask object metadata."""
        state = self._task.state()
        status = AWS_TASK_STATUS_MAP.get(state, JobStatus.UNKNOWN)
        self._cache_metadata["status"] = status
        return status

    def queue_position(self) -> Optional[int]:
        """Returns queue position from Braket QuantumTask.
        '>2000' returns as 2000 for typing consistency."""
        try:
            position = self._task.queue_position().queue_position
            if isinstance(position, str):
                if position.startswith(">"):
                    position = position[1:]
                return int(position)
            return position
        except AttributeError as err:
            raise AmazonBraketVersionError(
                "Queue visibility is only available for amazon-braket-sdk>=1.56.0"
            ) from err

    def result(self) -> Result:
        """Return the results of the job."""

        if not self.is_terminal_state():
            logger.info("Result will be available when the job has reached a final state.")

        bk_result = self._task.result()
        metadata = self._task.metadata()
        success = metadata["status"] == "COMPLETED"
        device_id = metadata["deviceArn"]
        job_id = metadata["quantumTaskArn"]

        result_mapping = {
            GateModelQuantumTaskResult: (BraketGateModelResultBuilder, GateModelResultData),
            AnalogHamiltonianSimulationQuantumTaskResult: (BraketAhsResultBuilder, AhsResultData),
        }

        builder_class, data_class = result_mapping.get(type(bk_result), (None, None))

        if not builder_class or not data_class:
            raise ValueError(f"Unsupported result type: {type(bk_result).__name__}")

        result_data = {
            "measurement_counts": builder_class(bk_result).get_counts() if success else None,
            "measurements": builder_class(bk_result).measurements() if success else None,
        }

        data = data_class(**result_data)
        return Result(device_id=device_id, job_id=job_id, success=success, data=data, **metadata)

    def cancel(self) -> None:
        """Cancel the quantum task."""
        task = self._task

        if self.is_terminal_state():
            raise JobStateError("Cannot cancel quantum job in terminal state.")

        try:
            task.cancel()
        except RuntimeError:
            task._aws_session.cancel_quantum_task(task.arn)

    @staticmethod
    def _get_cost(task_arn: str, aws_session: Optional[AwsSession] = None) -> Decimal:
        """Return the cost of the quantum task."""
        return get_quantum_task_cost(task_arn, aws_session=aws_session)

    def get_cost(self) -> float:
        """Return the cost of the job."""
        decimal_cost = self._get_cost(self._task.arn)
        return float(decimal_cost)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM conversion function

"""
import os
import textwrap

from qbraid._version import __version__ as qbraid_version
from qbraid.passes.qasm.decompose import _decompose_rxx_instr
from qbraid.passes.qasm.format import remove_unused_gates
from qbraid.programs.typer import Qasm2String, Qasm2StringType, Qasm3StringType
from qbraid.transpiler.annotations import weight


def _get_qasm3_gate_defs() -> str:
    """Helper function to get openqasm 3 gate defs from .qasm file"""
    current_dir = os.path.dirname(os.path.abspath(__file__))
    qelib_qasm = os.path.join(current_dir, "qelib_qasm3.qasm")
    with open(qelib_qasm, mode="r", encoding="utf-8") as file:
        gate_defs = file.read()
    return gate_defs


def _build_qasm_3_reg(line: str, qreg_type: bool) -> str:
    """Helper function to build openqasm 3 register statements

    Args:
        line (str): openqasm 2  regdecl statement
        qreg_type (bool): whether a qreg or creg type statement

    Returns:
        str : openqasm 3 qubits / bits declaration
    """
    reg_keyword_len = 4
    line = line[reg_keyword_len:]
    elements = line.split("[")
    reg_name = elements[0].strip()
    reg_size = elements[1].split("]")[0].strip()
    result = "\nqubit" if qreg_type else "bit"
    result += f"[{reg_size}] {reg_name};\n"
    return result


def _build_qasm_3_measure(line: str) -> str:
    """Helper function to build openqasm 3 measure string

    Args:
        line (str): openqasm 2 measure statement

    Returns:
        str:  openqasm 3 measure statement
    """
    measure_keyword_len = 7
    line = line[measure_keyword_len:]
    elements = line.split("->")
    qubits_name = elements[0].replace(" ", "")
    bits_name = elements[1].split(";")[0].replace(" ", "")

    return f"{bits_name} = measure {qubits_name};\n"


def _convert_line_to_qasm3(line: str) -> str:
    """Function to change an openqasm 2 line to openqasm 3

    Args:
        line (str): an openqasm 2 line

    Returns:
        str: corresponding openqasm 3 line
    """
    line = line.lstrip()
    if line.startswith("OPENQASM"):
        return ""
    if "qelib1.inc" in line:
        return ""
    if line.startswith("qreg"):
        return _build_qasm_3_reg(line, qreg_type=True)
    if line.startswith("creg"):
        return _build_qasm_3_reg(line, qreg_type=False)
    if line.startswith("u("):
        return line.replace("u(", "U(")
    if line.startswith("rxx("):
        return _decompose_rxx_instr(line)
    if line.startswith("measure"):
        return _build_qasm_3_measure(line)
    if line.startswith("opaque"):
        # as opaque is ignored by openqasm 3 add it as a comment
        return "// " + line + "\n"
    return line + "\n"


@weight(0.7)
def qasm2_to_qasm3(qasm_str: Qasm2StringType) -> Qasm3StringType:
    """Convert a OpenQASM 2.0 string to OpenQASM 3.0 string

    Args:
        qasm_str (str): OpenQASM 2.0 string

    Returns:
        str: OpenQASM 3.0 string
    """
    if not isinstance(qasm_str, Qasm2String):
        raise ValueError("Invalid OpenQASM 2.0 string")

    qasm3_str = textwrap.dedent(
        f"""
        // Generated from qBraid v{qbraid_version}
        OPENQASM 3.0;
        include 'stdgates.inc';
    """
    )

    gate_defs = _get_qasm3_gate_defs()

    for line in gate_defs:
        qasm3_str += line

    last_line_was_blank = False

    for line in qasm_str.splitlines():
        if line.strip().startswith("//"):
            continue

        # Check if the current line is blank
        if not line.strip():
            if last_line_was_blank:
                continue  # pragma: no cover

            last_line_was_blank = True
        else:
            last_line_was_blank = False

        line = _convert_line_to_qasm3(line)
        qasm3_str += line

    qasm3_str = remove_unused_gates(qasm3_str)

    return qasm3_str



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for providing transforamtions to ensure consistency
in the way OpenQASM 3 strings are formatted.

"""

import re
from typing import Optional


def _remove_empty_lines(input_string: str) -> str:
    """Removes all empty lines from the provided string."""
    return "\n".join(line for line in input_string.split("\n") if line.strip())


def _remove_double_empty_lines(qasm: str) -> str:
    """Replace double empty lines with single lines from a QASM string."""
    return re.sub(r"\n\n\n", "\n\n", qasm)


def _remove_comments(qasm: str, keep_header: bool = True) -> str:
    """Remove comments from a QASM string. Optionally keep the header."""
    lines = qasm.splitlines()
    parse_lines = lines[1:] if keep_header else lines
    result_lines = [lines[0]] + [line for line in parse_lines if not line.strip().startswith("//")]
    return "\n".join(result_lines)


def _format_qasm(qasm: str, skip_pattern: str) -> str:
    lines = qasm.split("\n")
    formatted_lines = []

    for line in lines:
        line = line.strip()  # Strip leading and trailing whitespace
        if skip_pattern.match(line) or line.startswith("//"):
            # If the line matches the gate definition pattern, add it as is
            formatted_lines.append(line)
        else:
            # Otherwise, split it at semicolons and add each part as a separate line
            parts = re.split(";[ ]*", line)
            parts = [part + ";" for part in parts if part]  # Remove empty parts
            formatted_lines.extend(parts)

    return "\n".join(formatted_lines)


def format_qasm(qasm: str, skip_pattern: Optional[str] = None) -> str:
    """Format a QASM string."""
    if skip_pattern is not None:
        return _format_qasm(qasm, skip_pattern)

    qasm = _remove_comments(qasm)
    qasm = _remove_double_empty_lines(qasm)
    return qasm.strip()


def _remove_gate_definition(qasm: str, gate_name: str) -> str:
    """Remove a gate definition from a QASM string."""
    lines = iter(qasm.split("\n"))
    new_qasm = ""

    for line in lines:
        if re.search(r"gate\s+(\w+)", line) is not None:
            # extract the gate name
            current_gate_name = re.search(r"gate\s+(\w+)", line).group(1)
            # remove lines from start curly brace to end curly brace
            if current_gate_name == gate_name:
                while "}" not in line:
                    line = next(lines)
            else:
                new_qasm += line + "\n"
        else:
            new_qasm += line + "\n"

    new_qasm = _remove_double_empty_lines(new_qasm)

    return new_qasm.strip()


def remove_unused_gates(qasm: str) -> str:
    """Remove unused gate definitions from a QASM string."""
    lines = iter(qasm.split("\n"))
    all_gates = {}

    for line in lines:
        if re.search(r"^\s*gate\s+(\w+)", line) is not None:
            gate_name = re.search(r"^\s*gate\s+(\w+)", line).group(1)
            all_gates[gate_name] = -1
        for gate in all_gates:
            if re.search(r"\b" + re.escape(gate) + r"\b", line):
                all_gates[gate] += 1

    new_qasm = qasm
    unused_gates = [gate for gate, count in all_gates.items() if count == 0]
    for gate in unused_gates:
        new_qasm = _remove_gate_definition(new_qasm, gate)

    if len(unused_gates) > 0:
        return remove_unused_gates(new_qasm)

    return new_qasm.strip()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
OpenQASM 3 conversions

.. currentmodule:: qbraid.transpiler.conversions.openqasm3

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   openqasm3_to_qasm3
   openqasm3_to_ionq

"""
from .openqasm3_to_ionq import openqasm3_to_ionq
from .openqasm3_to_qasm3 import openqasm3_to_qasm3

__all__ = ["openqasm3_to_qasm3", "openqasm3_to_ionq"]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
OpenQASM 2 conversions

.. currentmodule:: qbraid.transpiler.conversions.qasm2

Functions
----------

.. autosummary::
    :toctree: ../stubs/

    qasm2_to_cirq
    qasm2_to_pytket
    qasm2_to_qiskit
    qasm2_to_qasm3
    qasm2_to_ionq

"""
from .qasm2_to_cirq import qasm2_to_cirq
from .qasm2_to_ionq import qasm2_to_ionq
from .qasm2_to_pytket import qasm2_to_pytket
from .qasm2_to_qasm3 import qasm2_to_qasm3
from .qasm2_to_qiskit import qasm2_to_qiskit

__all__ = ["qasm2_to_cirq", "qasm2_to_pytket", "qasm2_to_qasm3", "qasm2_to_qiskit", "qasm2_to_ionq"]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing granular type checking for quantum programs
that use Python's built-in types.

"""
from abc import ABCMeta, abstractmethod
from typing import Any, Optional, Type, TypeVar

from openqasm3.parser import QASM3ParsingError, parse

from .exceptions import QasmError, ValidationError

IonQDictType = TypeVar("IonQDictType", bound=dict)


class QbraidMetaType(ABCMeta):
    """Abstract metaclass for custom program type checking based on built-in types."""

    @property
    @abstractmethod
    def __alias__(cls) -> Optional[str]:
        """The program type alias to associate with the class."""

    @property
    @abstractmethod
    def __bound__(cls) -> Type[Any]:
        """The built-in type the class is wrapping."""

    def __repr__(cls) -> str:
        return "~" + cls.__name__


class IonQDictInstanceMeta(QbraidMetaType):
    """Metaclass for IonQ JSON type checking based on dict content."""

    @property
    def __alias__(cls) -> str:
        return "ionq"

    @property
    def __bound__(cls) -> Type[dict]:
        return dict

    @staticmethod
    def _validate_field(single: Any, multiple: Any, field_name: str) -> None:
        """Helper method to validate single or multiple target/control fields."""
        if single is not None and multiple is not None:
            raise ValidationError(
                f"Both {field_name} and {field_name}s are set; only one should be provided."
            )
        if single is not None and not isinstance(single, int):
            raise ValidationError(f"Invalid {field_name}: {single}. Must be an integer.")
        if multiple is not None and not (
            isinstance(multiple, list) and all(isinstance(item, int) for item in multiple)
        ):
            raise ValidationError(f"Invalid {field_name}s: {multiple}. Must be a list of integers.")

    def __instancecheck__(cls, instance: Any) -> bool:
        """Custom instance checks based on dict format."""
        if not isinstance(instance, dict):
            return False

        qubits = instance.get("qubits")
        circuit = instance.get("circuit")
        gateset = instance.get("gateset")
        ciruit_format = instance.get("format")

        if ciruit_format is not None and not isinstance(ciruit_format, str):
            return False

        if gateset is not None and not isinstance(gateset, str):
            return False

        if not isinstance(qubits, int) or not isinstance(circuit, list):
            return False

        for op in circuit:
            if not isinstance(op, dict):
                return False

            gate = op.get("gate")
            rotation = op.get("rotation")
            target, targets = op.get("target"), op.get("targets")
            control, controls = op.get("control"), op.get("controls")

            if not isinstance(gate, str):
                return False

            if rotation is not None and not isinstance(rotation, (int, float)):
                return False

            try:
                cls._validate_field(target, targets, "target")
                cls._validate_field(control, controls, "control")
            except ValidationError:
                return False

        return True


class IonQDict(metaclass=IonQDictInstanceMeta):
    """Marker class for dict that are valid IonQ JSON formatted programs."""


def extract_qasm_version(qasm: str) -> int:
    """
    Parses an OpenQASM program string to determine its version, either 2 or 3.

    Args:
        qasm (str): The OpenQASM program string.

    Returns:
        int: The OpenQASM version as an integer.

    Raises:
        QasmError: If the string does not represent a valid OpenQASM program.
    """
    qasm = qasm.replace("opaque", "// opaque")  # Temporarily mask out the opaque keyword
    try:
        parsed_program = parse(qasm)
        version = int(float(parsed_program.version))
        return version
    except (QASM3ParsingError, ValueError, TypeError) as err:
        raise QasmError("Could not determine the OpenQASM version.") from err


def get_qasm_type_alias(qasm: str) -> str:
    """
    Determines the type alias for an OpenQASM program based on its version.

    Args:
        qasm (str): The OpenQASM program string.

    Returns:
        str: The QASM version alias ('qasm2' or 'qasm3').

    Raises:
        QasmError: If the string does not represent a valid OpenQASM program.
    """
    try:
        version = extract_qasm_version(qasm)
        type_alias = f"qasm{version}"
        return type_alias
    except QasmError as err:
        raise QasmError(
            "Could not determine the type alias: the OpenQASM program may be invalid."
        ) from err


class BaseQasmInstanceMeta(QbraidMetaType):
    """Metaclass for OpenQASM type checking based on string content.

    Attributes:
        version (Optional[int]): The specific OpenQASM type to check for.
    """

    version: Optional[int] = None

    @property
    def __alias__(cls) -> Optional[str]:
        if isinstance(cls.version, int):
            return f"qasm{cls.version}"
        return None

    @property
    def __bound__(cls) -> Type[str]:
        return str

    def __instancecheck__(cls, instance: Any) -> bool:
        """Custom instance checks based on OpenQASM type.

        Args:
            instance: The object to check.

        Returns:
            bool: True if instance is a string matching the OpenQASM type, False otherwise.
        """
        if not isinstance(instance, str):
            return False
        try:
            return extract_qasm_version(instance) == cls.version
        except QasmError:
            return False


class Qasm2StringMeta(BaseQasmInstanceMeta):
    """Metaclass for instances representing OpenQASM 2 strings."""

    version = 2


class Qasm3StringMeta(BaseQasmInstanceMeta):
    """Metaclass for instances representing OpenQASM 3 strings."""

    version = 3


class Qasm2String(metaclass=Qasm2StringMeta):
    """Marker class for strings that are valid OpenQASM 2 programs."""


class Qasm3String(metaclass=Qasm3StringMeta):
    """Marker class for strings that are valid OpenQASM 3 programs."""


class QasmStringType(str):
    """Base class for OpenQASM string types, providing validation upon instantiation."""

    version: Optional[int] = None

    def __new__(cls, value):
        if not isinstance(value, str):
            raise TypeError("OpenQASM strings must be initialized with a string.")
        if not extract_qasm_version(value) == cls.version:
            raise ValueError(f"String does not conform to OpenQASM {cls.version} format.")
        return str.__new__(cls, value)


class Qasm2StringType(QasmStringType):
    """Specifically typed string for OpenQASM 2 formatted text."""

    version = 2


class Qasm3StringType(QasmStringType):
    """Specifically typed string for OpenQASM 3 formatted text."""

    version = 3


class QuboCoefficientsDictInstanceMeta(QbraidMetaType):
    """Metaclass for Qubo coefficients JSON type checking based on dict content."""

    @property
    def __alias__(cls) -> str:
        return "qubo"

    @property
    def __bound__(cls) -> Type[dict]:
        return dict

    def __instancecheck__(cls, instance: Any) -> bool:
        """Custom instance checks based on dict format."""
        if not instance or not isinstance(instance, dict):
            return False

        for key, value in instance.items():
            if not (
                isinstance(key, tuple) and len(key) == 2 and all(isinstance(k, str) for k in key)
            ):
                return False

            if not isinstance(value, (float, int)):
                return False

        return True


class QuboCoefficientsDict(metaclass=QuboCoefficientsDictInstanceMeta):
    """Marker class for dict that are valid Qubo coefficients format."""


QBRAID_META_TYPES = {IonQDict, QuboCoefficientsDict}
BOUND_QBRAID_META_TYPES = {Qasm2String, Qasm3String}



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for preprocessing qasm string to before it is passed to parser.

"""
import re

from .compat import remove_qasm_barriers
from .decompose import decompose_qasm2
from .format import format_qasm


def _unfold_gate_defs(qasm: str) -> str:
    """Recursively expands gate definitions in the input OpenQASM string."""
    # Define regular expression patterns
    gate_definition_pattern = re.compile(
        r"gate ([a-zA-Z0-9_]+)(\((.*?)\))? ((q[0-9]+,)*q[0-9]+) {(.*?)}"
    )

    gate_body = ""
    params_list = []
    gate_usage_match = None

    # Find gate definition and extract its components
    gate_definition_match = gate_definition_pattern.search(qasm)
    if gate_definition_match:
        gate_name, _, params, qubits, _, gate_body = gate_definition_match.groups()
        params_list = [param.strip() for param in params.split(",")] if params is not None else []

        qubits = [qubit.strip() for qubit in qubits.split(",")]

        # pylint: disable=consider-using-f-string
        gate_usage_pattern = re.compile(
            r"({})(\((.*?)\))? ((q\[([0-9]+)\],)*(q\[([0-9]+)\]));".format(gate_name)
        )

        # Replace parameters with their values in gate body
        gate_usage_match = gate_usage_pattern.search(qasm)

    while gate_usage_match:
        groups = gate_usage_match.groups()
        param_values, qubits_usage = groups[2], groups[3]
        param_values_list = (
            [value.strip() for value in param_values.split(",")] if param_values is not None else []
        )
        expanded_gate_body = gate_body
        qubits_usage = [qubit.strip() for qubit in re.findall(r"q\[\d+\]", qubits_usage)]

        for param, value in zip(params_list, param_values_list):
            expanded_gate_body = expanded_gate_body.replace(param, value)

        for qubit, qubit_usage in zip(qubits, qubits_usage):
            expanded_gate_body = expanded_gate_body.replace(qubit, qubit_usage)

        # Replace gate usage with the expanded gate body in the input string
        qasm = qasm.replace(gate_usage_match.group(0), expanded_gate_body + ";")

        # Search for the next gate usage
        gate_usage_match = gate_usage_pattern.search(qasm)

    # Remove double semicolons
    qasm = format_qasm(qasm, gate_definition_pattern)

    return qasm


def _find_gate_line(lines):
    for i, line in enumerate(lines):
        if line.strip().startswith("gate"):
            return i
    return None


def unfold_qasm2(qasm: str) -> str:
    """Returns a QASM copy with custom gates deconstructed and unsupported operations decomposed."""
    # temp hack to fix 'r' replacing last char of 'ecr'
    qasm = qasm.replace("ecr", "ecr_")

    input_str = remove_qasm_barriers(qasm)

    lines = input_str.strip("\n").split("\n")
    gate_lines = [(i, line) for i, line in enumerate(lines) if line.strip().startswith("gate")]
    gate_lines.reverse()  # Reverse to start removing from the last
    gate_line_idx = _find_gate_line(lines)

    # Remove all 'gate' lines
    for idx, _ in gate_lines:
        lines.pop(idx)

    for _, gate_line in gate_lines:
        # Insert the current 'gate' line for this iteration
        lines.insert(gate_line_idx, gate_line)

        new_input = "\n".join(lines)
        new_input = _unfold_gate_defs(new_input)  # call the conversion function
        lines = new_input.split("\n")  # update lines after conversion

        # Remove the current 'gate' line for the next iteration
        lines.pop(gate_line_idx)

    qasm = "\n".join(lines)
    qasm_out = decompose_qasm2(qasm)

    return qasm_out



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining IonQ device class

"""
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Optional, Union

from qbraid.programs import load_program
from qbraid.programs.typer import IonQDict, IonQDictType, QasmStringType
from qbraid.runtime.device import QuantumDevice
from qbraid.runtime.enums import DeviceStatus
from qbraid.transpiler.conversions.openqasm3.openqasm3_to_ionq import (
    IONQ_ONE_QUBIT_GATE_MAP,
    IONQ_THREE_QUBIT_GATE_MAP,
    IONQ_TWO_QUBIT_GATE_MAP,
)

from .job import IonQJob

if TYPE_CHECKING:
    import qbraid.runtime
    import qbraid.runtime.ionq.provider


IONQ_GATE_MAP = IONQ_ONE_QUBIT_GATE_MAP | IONQ_TWO_QUBIT_GATE_MAP | IONQ_THREE_QUBIT_GATE_MAP


class IonQDevice(QuantumDevice):
    """IonQ quantum device interface."""

    def __init__(
        self,
        profile: qbraid.runtime.TargetProfile,
        session: qbraid.runtime.ionq.provider.IonQSession,
    ):
        super().__init__(profile=profile)
        self._session = session

    @property
    def session(self) -> qbraid.runtime.ionq.provider.IonQSession:
        """Return the IonQ session."""
        return self._session

    def __str__(self):
        """String representation of the IonQDevice object."""
        return f"{self.__class__.__name__}('{self.id}')"

    def status(self) -> qbraid.runtime.DeviceStatus:
        """Return the current status of the IonQ device."""
        device_data = self.session.get_device(self.id)
        status = device_data.get("status")

        if status in ["available", "running"]:
            return DeviceStatus.ONLINE

        if status in ["unavailable", "reserved", "calibrating"]:
            return DeviceStatus.UNAVAILABLE

        if status == "retired":
            return DeviceStatus.RETIRED

        if status == "offline":
            return DeviceStatus.OFFLINE

        raise ValueError(f"Unrecognized device status: {status}")

    def avg_queue_time(self) -> int:
        """Return the average queue time for the IonQ device."""
        device_data = self.session.get_device(self.id)
        return device_data["average_queue_time"]

    def transform(self, run_input: QasmStringType) -> QasmStringType:
        """Transform the input to the IonQ device."""
        program = load_program(run_input)
        program.transform(device=self, gate_mappings=IONQ_GATE_MAP)
        return program.program

    @staticmethod
    def _squash_multicircuit_input(batch_input: list[IonQDictType]) -> dict[str, Any]:
        if not batch_input:
            raise ValueError("run_input list cannot be empty.")

        default_format = "ionq.circuit.v0"
        default_gateset = "qis"

        input_format = batch_input[0].get("format", default_format)
        input_gateset = batch_input[0].get("gateset", default_gateset)
        max_qubits = 0
        circuits = []

        for i, run_input in enumerate(batch_input):
            if not isinstance(run_input, IonQDict):
                raise ValueError("All run_inputs must be an instance of ~IonQDict.")
            if run_input.get("format", default_format) != input_format:
                raise ValueError("All run_inputs must have the same value for key 'format'.")
            if run_input.get("gateset", default_gateset) != input_gateset:
                raise ValueError("All run_inputs must have the same value for key 'gateset'.")

            max_qubits = max(max_qubits, run_input["qubits"])
            circuits.append({"circuit": run_input["circuit"], "name": f"Circuit {i}"})

        return {
            "format": input_format,
            "gateset": input_gateset,
            "qubits": max_qubits,
            "circuits": circuits,
        }

    # pylint:disable-next=arguments-differ,too-many-arguments
    def submit(
        self,
        run_input: Union[IonQDictType, list[IonQDictType]],
        shots: int,
        preflight: bool = False,
        name: Optional[str] = None,
        noise: Optional[dict[str, Any]] = None,
        error_mitigation: Optional[dict[str, Any]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs,
    ) -> IonQJob:
        """Submit a job to the IonQ device."""
        ionq_input = (
            self._squash_multicircuit_input(run_input) if isinstance(run_input, list) else run_input
        )
        job_data = {
            "target": self.id,
            "shots": shots,
            "preflight": preflight,
            "input": ionq_input,
            **kwargs,
        }
        optional_fields = {
            "name": name,
            "noise": noise,
            "metadata": metadata,
            "error_mitigation": error_mitigation,
        }
        job_data.update({key: value for key, value in optional_fields.items() if value is not None})
        serialized_data = json.dumps(job_data)
        job_data = self.session.create_job(serialized_data)
        job_id = job_data.get("id")
        if not job_id:
            raise ValueError("Job ID not found in the response")
        return IonQJob(job_id=job_id, session=self.session, device=self, shots=shots)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining BraketCircuit Class

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from braket.circuits import Circuit, Instruction, Qubit

from qbraid.programs.exceptions import ProgramTypeError

from ._model import GateModelProgram

if TYPE_CHECKING:
    import braket.circuits
    import numpy as np


class BraketCircuit(GateModelProgram):
    """Wrapper class for ``braket.circuits.Circuit`` objects."""

    def __init__(self, program: braket.circuits.Circuit):
        super().__init__(program)
        if not isinstance(program, Circuit):
            raise ProgramTypeError(
                message=f"Expected 'braket.circuits.Circuit' object, got '{type(program)}'."
            )

    @property
    def qubits(self) -> list[Qubit]:
        """Return the qubits acted upon by the operations in this circuit"""
        return list(self.program.qubits)

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        return self.program.qubit_count

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return 0

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return self.program.depth

    def _unitary(self) -> "np.ndarray":
        """Calculate unitary of circuit."""
        return self.program.to_unitary()

    def populate_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, adds identity gates to vacant registers as needed."""
        max_qubit = 0
        occupied_qubits = []
        circuit = self.program.copy()
        for qubit in circuit.qubits:
            index = int(qubit)
            occupied_qubits.append(index)
            max_qubit = max(max_qubit, index)
        qubit_count = max_qubit + 1
        if qubit_count > circuit.qubit_count:
            all_qubits = list(range(0, qubit_count))
            vacant_qubits = list(set(all_qubits) - set(occupied_qubits))
            for index in vacant_qubits:
                circuit.i(index)
        self._program = circuit

    def remove_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, reduces dimension accordingly."""
        qubit_map = {}
        circuit = self.program.copy()
        circuit_qubits = list(circuit.qubits)
        circuit_qubits.sort()
        for index, qubit in enumerate(circuit_qubits):
            qubit_map[int(qubit)] = index
        contig_circuit = Circuit()
        for instr in circuit.instructions:
            contig_target = [qubit_map[int(qubit)] for qubit in list(instr.target)]
            contig_control = [qubit_map[int(qubit)] for qubit in list(instr.control)]
            contig_instr = Instruction(
                instr.operator,
                target=contig_target,
                control=contig_control,
                control_state=instr.control_state,
                power=instr.power,
            )
            contig_circuit.add_instruction(contig_instr)
        self._program = contig_circuit

    def reverse_qubit_order(self) -> None:
        """Reverse the order of the qubits in the circuit."""
        qubit_map = {}
        circuit = self.program.copy()
        circuit_qubits = list(circuit.qubits)
        circuit_qubits.sort()
        circuit_qubits = list(reversed(circuit_qubits))
        for index, qubit in enumerate(circuit_qubits):
            qubit_map[int(qubit)] = index
        contig_circuit = Circuit()
        for instr in circuit.instructions:
            contig_target = [qubit_map[int(qubit)] for qubit in list(instr.target)]
            contig_control = [qubit_map[int(qubit)] for qubit in list(instr.control)]
            contig_instr = Instruction(
                instr.operator,
                target=contig_target,
                control=contig_control,
                control_state=instr.control_state,
                power=instr.power,
            )
            contig_circuit.add_instruction(contig_instr)
        self._program = contig_circuit

    def transform(self, device) -> None:
        """Transform program to according to device target profile."""
        if device.simulator:
            self.remove_idle_qubits()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Functions for analyzing OpenQASM programs.

"""
from __future__ import annotations

from typing import Optional, Union

from openqasm3.ast import (
    BinaryExpression,
    BitType,
    BranchingStatement,
    ClassicalDeclaration,
    Concatenation,
    Expression,
    Identifier,
    IndexedIdentifier,
    IntegerLiteral,
    Program,
    QuantumBarrier,
    QuantumGate,
    QuantumMeasurementStatement,
    QuantumReset,
    QubitDeclaration,
    RangeDefinition,
    Statement,
)
from openqasm3.parser import parse


def has_measurements(program: Union[Program, str]) -> bool:
    """Check if the program has any measurement operations."""
    program = parse(program) if isinstance(program, str) else program
    for statement in program.statements:
        if isinstance(statement, QuantumMeasurementStatement):
            return True
    return False


def expression_value(expression: Optional[Union[Expression, RangeDefinition]]) -> int:
    """Return the size of an expression."""
    if isinstance(expression, IntegerLiteral):
        return expression.value

    raise ValueError(f"Invalid expression type: {type(expression)}. Expected IntegerLiteral.")


def expression_value_option(expression: Optional[Expression]) -> Optional[int]:
    """Return the size of an expression."""
    if expression is None:
        return None

    return expression_value(expression)


# pylint: disable-next=too-many-statements
def depth(
    qasm_statements: list[Statement], counts: dict[tuple[str, int], int]
) -> dict[tuple[str, int], int]:
    """Return the depth of a list of given qasm statements."""
    qreg_sizes = {}
    creg_sizes = {}
    track_measured = {}
    max_depth = 0
    # pylint: disable-next=too-many-nested-blocks
    for statement in qasm_statements:
        if isinstance(statement, QubitDeclaration):
            qreg_name = statement.qubit.name
            qreg_size = expression_value(statement.size)
            qreg_sizes[qreg_name] = qreg_size
            continue
        if isinstance(statement, ClassicalDeclaration) and isinstance(statement.type, BitType):
            creg_name = statement.identifier.name
            creg_size: int = expression_value(statement.type.size)
            creg_sizes[creg_name] = creg_size
            for i in range(creg_size):
                track_measured[f"{creg_name}[{i}]"] = 0
            continue
        if isinstance(statement, QuantumGate):
            qubits_involved = set()
            if all(isinstance(qubit, IndexedIdentifier) for qubit in statement.qubits):
                for qubit in statement.qubits:
                    if isinstance(qubit.name, Identifier):
                        qreg_name = qubit.name.name
                        if isinstance(qubit.indices[0], list):
                            expression = qubit.indices[0][0]
                        qubit_index = expression_value(expression)
                        counts[(qreg_name, qubit_index)] += 1
                        qubits_involved.add((qreg_name, qubit_index))
                max_involved_depth = max(counts[qubit] for qubit in qubits_involved)
                for qubit in qubits_involved:
                    counts[qubit] = max_involved_depth
            else:
                for qubit in statement.qubits:
                    qreg_name = str(qubit.name)
                    for i in range(qreg_sizes[qreg_name]):
                        counts[(qreg_name, i)] += 1
            max_depth = max(counts.values())
        elif isinstance(statement, QuantumReset):
            if isinstance(statement.qubits, IndexedIdentifier):
                qreg_name = statement.qubits.name.name
                if isinstance(statement.qubits.indices[0], list):
                    expression = statement.qubits.indices[0][0]
                qubit_index = expression_value(expression)
                counts[(qreg_name, qubit_index)] += 1
            else:
                qreg_name = statement.qubits.name
                for i in range(qreg_sizes[qreg_name]):
                    counts[(qreg_name, i)] += 1
        elif isinstance(statement, QuantumBarrier):
            for qubit_identifier in statement.qubits:
                if isinstance(qubit_identifier, (IndexedIdentifier, Identifier)):
                    qreg_name = str(qubit_identifier.name)
                    for i in range(qreg_sizes[qreg_name]):
                        counts[(qreg_name, i)] = max_depth
        elif isinstance(statement, QuantumMeasurementStatement):
            qubit = statement.measure.qubit
            if isinstance(qubit, IndexedIdentifier):
                qreg_name = qubit.name.name
                if isinstance(qubit.indices[0], list):
                    qubit_expr = qubit.indices[0][0]
                qubit_index = expression_value(qubit_expr)
                counts[(qreg_name, qubit_index)] += 1
                max_depth = max(counts.values())
                if isinstance(statement.target, IndexedIdentifier):
                    if isinstance(statement.target.indices[0], list):
                        creg_expr = statement.target.indices[0][0]
                        creg_index = expression_value(creg_expr)
                        creg_name = statement.target.name.name
                        track_measured[(creg_name, creg_index)] = max_depth
            else:
                qreg_name = qubit.name
                for i in range(qreg_sizes[qreg_name]):
                    counts[(qreg_name, i)] += 1
                if isinstance(statement.target, Identifier):
                    creg = str(statement.target.name)
                max_depth = max(counts.values())
                for i in range(creg_sizes[creg]):
                    track_measured[(creg, i)] = max_depth
        elif isinstance(statement, BranchingStatement) and isinstance(
            statement.condition, (BinaryExpression, Concatenation)
        ):
            expression = statement.condition.lhs
            if isinstance(expression, (IndexedIdentifier, Identifier)):
                creg_name = expression.name

                for creg_index in range(creg_sizes[creg_name]):
                    if (creg_name, creg_index) not in track_measured:
                        track_measured[(creg_name, creg_index)] = 0

                required_depth = max(
                    track_measured[(creg_name, creg_index)]
                    for creg_index in range(creg_sizes[creg_name])
                )
                required_depth = max(required_depth, max_depth)
                for i in range(creg_sizes[creg_name]):
                    track_measured[(creg_name, i)] = required_depth
                qubits: set[str] = set()
                for sub_statement in statement.if_block + statement.else_block:
                    if isinstance(sub_statement, QuantumGate):
                        for qubit in sub_statement.qubits:
                            if isinstance(qubit.name, Identifier):
                                qreg_name = qubit.name.name
                            if isinstance(qubit, IndexedIdentifier):
                                if isinstance(qubit.indices[0], list):
                                    expression = qubit.indices[0][0]
                            qubit_index = expression_value(expression)
                            qubits.add((qreg_name, qubit_index))
                    elif isinstance(sub_statement, QuantumMeasurementStatement):
                        if isinstance(sub_statement.measure.qubit.name, Identifier):
                            qreg_name = sub_statement.measure.qubit.name.name
                        if isinstance(sub_statement.measure.qubit, IndexedIdentifier):
                            if isinstance(sub_statement.measure.qubit.indices[0], list):
                                expression = sub_statement.measure.qubit.indices[0][0]
                        if isinstance(expression, Expression):
                            qubit_index = expression_value(expression)
                            qubits.add((qreg_name, qubit_index))

                for qubit_id in qubits:
                    counts[qubit_id] = max(required_depth, counts[qubit_id]) + 1

                max_depth = max(counts.values())
    return counts



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining BraketGateModelResultBuilder Class

"""
from __future__ import annotations

from collections import Counter
from typing import TYPE_CHECKING, Optional

import numpy as np

from qbraid.runtime.exceptions import QbraidRuntimeError
from qbraid.runtime.result_data import AhsShotResult

if TYPE_CHECKING:
    from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import (
        AnalogHamiltonianSimulationQuantumTaskResult,
    )
    from braket.tasks.gate_model_quantum_task_result import GateModelQuantumTaskResult


class ResultDecodingError(QbraidRuntimeError):
    """Exception raised for errors that occur during the decoding of result data."""


class BraketGateModelResultBuilder:
    """Wrapper class for Amazon Braket result objects."""

    def __init__(self, result: GateModelQuantumTaskResult):
        self._result = result

    def measurements(self) -> np.ndarray:
        """
        2d array - row is shot and column is qubit. Default is None.
        Only available when shots > 0. The qubits in `measurements` are
        the ones in `GateModelQuantumTaskResult.measured_qubits`.

        """
        result: GateModelQuantumTaskResult = self._result
        return np.flip(result.measurements, 1)

    def get_counts(self) -> dict[str, int]:
        """Returns the histogram data of the run"""
        result: GateModelQuantumTaskResult = self._result
        braket_counts = dict(result.measurement_counts)
        qbraid_counts = {}
        for key in braket_counts:
            str_key = "".join(reversed([str(i) for i in key]))
            qbraid_counts[str_key] = braket_counts[key]
        return qbraid_counts


class BraketAhsResultBuilder:
    """Result from an Analog Hamiltonian Simulation (AHS)."""

    def __init__(self, result: AnalogHamiltonianSimulationQuantumTaskResult):
        self._result = result

    def measurements(self) -> list[AhsShotResult]:
        """Get the list of shot results from the AHS job."""
        result: AnalogHamiltonianSimulationQuantumTaskResult = self._result
        return [
            AhsShotResult(
                success=m.status.name == "SUCCESS",
                pre_sequence=m.pre_sequence,
                post_sequence=m.post_sequence,
            )
            for m in result.measurements
        ]

    def get_counts(self) -> Optional[dict[str, int]]:
        """
        Aggregate state counts from AHS shot results.

        This function decodes the state of atoms at different sites based on
        pre- and post-measurement sequences. Each atom can be in one of three states:
        empty site ('e'), Rydberg state ('r'), or ground state ('g').

        Returns:
            Optional[dict[str, int]]: A dictionary mapping each unique state configuration to
                                      the count of its occurrences across all successful shots,
                                      or None if there are no successful measurements.

        Raises:
            ResultDecodingError: If there is an error accessing required attributes from
                the result object.

        """
        state_counts = Counter()
        states = ["e", "r", "g"]
        try:
            for shot in self.measurements():
                if shot.success:
                    pre = shot.pre_sequence
                    post = shot.post_sequence
                    # converting presequence and postsequence measurements to state_idx
                    state_idx = [
                        0 if pre_i == 0 else 1 if post_i == 0 else 2
                        for pre_i, post_i in zip(pre, post)
                    ]
                    state = "".join(states[s_idx] for s_idx in state_idx)
                    state_counts.update([state])
        except AttributeError as err:
            raise ResultDecodingError from err

        return None if not state_counts else dict(state_counts)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining OpenQasm2Program class.

"""
from __future__ import annotations

import re
from typing import TYPE_CHECKING

import numpy as np
from openqasm3.ast import Program
from openqasm3.parser import parse
from qbraid_core._import import LazyLoader

from qbraid.passes.qasm import depth, normalize_qasm_gate_params, rebase, remove_measurements
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.typer import Qasm2String, Qasm2StringType

from ._model import GateModelProgram

if TYPE_CHECKING:
    import qbraid.runtime

transpiler = LazyLoader("transpiler", globals(), "qbraid.transpiler")


class OpenQasm2Program(GateModelProgram):
    """Wrapper class for OpenQASM 2 strings."""

    def __init__(self, program: Qasm2StringType):
        super().__init__(program)
        if not isinstance(program, Qasm2String):
            raise ProgramTypeError(message=f"Expected 'str' object, got '{type(program)}'.")

    def parsed(self) -> Program:
        """Parse the program string."""
        return parse(self._program)

    def _get_bits(self, bit_type: str) -> list[tuple[str, int]]:
        """Return the number of qubits or classical bits in the circuit.

        Args:
            bit_type: either "q" or "c" for qubits or classical bits, respectively.

        Returns:
            list[tuple[str, int]]: A list of qubits or classical bits. Each bit is
                a tuple of (register name, index)
        """
        matches = re.findall(rf"{bit_type}reg (\w+)\[(\d+)\];", self.program)

        result = []
        for match in matches:
            var_name = match[0]
            n = int(match[1])
            result.extend([(var_name, i) for i in range(n)])
        return result

    @property
    def qubits(self) -> list[tuple[str, int]]:
        """Use regex to extract all qreg definitions from the string"""
        return self._get_bits("q")

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return len(self._get_bits("c"))

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        counts = {q: 0 for q in self.qubits}
        program = self.parsed()
        counts = depth(program.statements, counts)
        return max(counts.values())

    def _unitary(self) -> np.ndarray:
        """Return the unitary of the QASM"""
        return transpiler.transpile(self.program, "cirq").unitary()

    def transform(self, device: qbraid.runtime.QuantumDevice, **kwargs) -> None:
        """Transform program to according to device target profile."""
        if device.id == "quera_qasm_simulator":
            self._program = remove_measurements(self.program)

        basis_gates = device.profile.get("basis_gates")

        if basis_gates is not None and len(basis_gates) > 0:
            transformed_qasm = rebase(self.program, basis_gates, **kwargs)
            self._program = normalize_qasm_gate_params(transformed_qasm)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing functions to convert between OpenQASM 3 and IonQ JSON format.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid._logging import logger
from qbraid.programs.gate_model.ionq import IONQ_NATIVE_GATES
from qbraid.transpiler.annotations import weight
from qbraid.transpiler.conversions.openqasm3.openqasm3_to_ionq import openqasm3_to_ionq
from qbraid.transpiler.exceptions import ProgramConversionError

pyqasm = LazyLoader("pyqasm", globals(), "pyqasm")

if TYPE_CHECKING:
    from qbraid.programs.typer import IonQDictType, Qasm3StringType


@weight(1)
def qasm3_to_ionq(qasm: Qasm3StringType) -> IonQDictType:
    """
    Convert an OpenQASM 3 string to IonQ JSON format.

    Args:
        qasm (Qasm3StringType): The OpenQASM 3 string to convert.

    Returns:
        IonQDictType: IonQ JSON format equivalent to the input OpenQASM 3 string.

    Raises:
        ProgramConversionError: If the conversion fails.
    """
    # pylint: disable=broad-exception-caught
    try:
        return openqasm3_to_ionq(qasm)
    except Exception as err:
        err_msg = str(err) or "Failed to convert OpenQASM 3 to IonQ JSON."
        cache_err = None
        try:
            if not any(gate in qasm for gate in IONQ_NATIVE_GATES):
                module = pyqasm.load(qasm)
                module.unroll()
                return openqasm3_to_ionq(module.unrolled_qasm)
        except (ImportError, ModuleNotFoundError) as import_err:
            cache_err = import_err
        except Exception as pyqasm_err:  # pragma: no cover
            logger.debug("Conversion with pyqasm assistance failed: %s", pyqasm_err)
        if (
            cache_err
            and "Cannot mix native and QIS gates in the same circuit" not in err_msg
            and "Circuits with measurements are not supported by the IonQDictType" not in err_msg
        ):
            err_msg += "." if err_msg and not err_msg.endswith(".") else ""
            err_msg += "" if not err_msg else " "
            err_msg += "Please install the 'ionq' extra to enable program unrolling with pyqasm."
        else:
            cache_err = err
        raise ProgramConversionError(err_msg) from cache_err
    # pylint: enable=broad-exception-caught



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM 2 conversion extras.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import requires_extras

qibo = LazyLoader("qibo", globals(), "qibo")

if TYPE_CHECKING:
    import qibo as qibo_  # type: ignore

    from qbraid.programs.typer import Qasm2StringType


@requires_extras("qibo")
def qasm2_to_qibo(qasm: Qasm2StringType) -> qibo_.Circuit:
    """Returns a qibo.Circuit equivalent to the input OpenQASM 2 circuit.

    Args:
        qasm: OpenQASM 2 string to convert to a qibo.Circuit

    Returns:
        qibo.Circuit object equivalent to the input OpenQASM 2 string.
    """
    # Remove problematic comment lines in the qasm code
    lines = [
        line.replace(", ", ",") for line in qasm.split("\n") if not line.strip().startswith("//")
    ]
    # Remove in line problematic comments
    clean_lines = []
    for line in lines:
        clean_line = line.split("//")[0].strip()
        if clean_line:
            clean_lines.append(clean_line)
    qasm = "\n".join(clean_lines)

    return qibo.Circuit.from_qasm(qasm)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining OpenQasm3Program class.

"""
from __future__ import annotations

import re
from typing import TYPE_CHECKING, Optional

import numpy as np
from openqasm3.ast import BitType, ClassicalDeclaration, Program, QubitDeclaration
from openqasm3.parser import parse

from qbraid.passes.qasm import depth, normalize_qasm_gate_params, rebase
from qbraid.passes.qasm.analyze import expression_value_option
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.typer import Qasm3String, Qasm3StringType

from ._model import GateModelProgram

if TYPE_CHECKING:
    import qbraid.runtime


def auto_reparse(func):
    """Decorator that ensures the quantum circuit's state
    is reparsed from QASM after method execution."""

    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        self._parse_state()
        return result

    return wrapper


class OpenQasm3Program(GateModelProgram):
    """Wrapper class for OpenQASM 3 strings."""

    def __init__(self, program: Qasm3StringType):
        super().__init__(program)
        if not isinstance(program, Qasm3String):
            raise ProgramTypeError(message=f"Expected 'str' object, got '{type(program)}'.")
        self._program: str = program
        self._parse_state()

    def parsed(self) -> Program:
        """Parse the program string."""
        return parse(self._program)

    def _parse_state(self) -> None:
        """Process the program string."""
        program = self.parsed()

        num_qubits = 0
        num_clbits = 0
        qubits: list[tuple[str, Optional[int]]] = []
        clbits: list[tuple[str, Optional[int]]] = []

        for statement in program.statements:
            if isinstance(statement, QubitDeclaration):
                name = statement.qubit.name
                size = expression_value_option(statement.size)
                qubits.append((name, size))
                num_qubits += 1 if size is None else size
            elif isinstance(statement, ClassicalDeclaration) and isinstance(
                statement.type, BitType
            ):
                name = statement.identifier.name
                size = expression_value_option(statement.type.size)
                clbits.append((name, size))
                num_clbits += 1 if size is None else size

        self._num_qubits = num_qubits
        self._num_clbits = num_clbits
        self._qubits = qubits
        self._clbits = clbits

    @property
    def qubits(self) -> list[tuple[str, Optional[int]]]:
        """Return the qubits acted upon by the operations in this circuit"""
        return self._qubits

    @property
    def clbits(self) -> list[tuple[str, Optional[int]]]:
        """Return the qubits acted upon by the operations in this circuit"""
        return self._clbits

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        return self._num_qubits

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return self._num_clbits

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        counts = {(q, i): 0 for q, s in self.qubits for i in range(s or 1)}
        program = self.parsed()
        counts = depth(program.statements, counts)
        return max(counts.values())

    def _unitary(self) -> np.ndarray:
        """Calculate unitary of circuit."""
        raise NotImplementedError

    @staticmethod
    def _remove_gate_definitions(qasm_str: str) -> str:
        """This is required to account for the case when the gate
        definition has an argument which is having same name as a
        quantum register

        now, if any gate is applied on this argument, it will be
        interpreted as being applied on THE WHOLE register, when it is
        only applied on the argument.

        Example :

        gate custom q1 {
            x q1; // this is STILL DETECTED as a gate application on q1
        }
        qreg q1[4];
        qreg q2[2];
        custom q1[0];
        cx q1[1], q2[1];

        // Actual depth : 1
        // Calculated depth : 2 (because of the gate definition)

        Args:
            qasm_str (string): The qasm string
        Returns:
            qasm_str (string): The qasm string with gate definitions removed
        """
        gate_decls = [x.group() for x in re.finditer(r"(gate)(.*\n)*?\s*\}", qasm_str)]
        for decl in gate_decls:
            qasm_str = qasm_str.replace(decl, "")
        return qasm_str

    def _get_unused_qubit_indices(self) -> dict:
        """Get unused qubit indices in the circuit

        Returns:
            dict: A dictionary with keys as register names and values as sets of unused indices
        """
        qasm_str = self._remove_gate_definitions(self.program)
        lines = qasm_str.splitlines()
        gate_lines = [
            s
            for s in lines
            if s.strip()
            and not s.strip().startswith(("OPENQASM", "include", "qreg", "qubit", "bit", "//"))
        ]
        unused_indices = {}
        for qreg, size in self.qubits:
            size = 1 if size is None else size
            unused_indices[qreg] = set(range(size))

            for line in gate_lines:
                if qreg not in line:
                    continue
                # either qubits or full register is referenced
                used_indices = {int(x) for x in re.findall(rf"{qreg}\[(\d+)\]", line)}
                if len(used_indices) > 0:
                    unused_indices[qreg] = unused_indices[qreg].difference(used_indices)
                else:
                    # full register is referenced
                    unused_indices[qreg] = set()
                    break

                if len(unused_indices[qreg]) == 0:
                    break

        return unused_indices

    @staticmethod
    def _remap_qubits(qasm_str, reg_name, reg_size, unused_indices):
        """Re-map the qubits for a partially used quantum register
        Args:
            qasm_str (str): QASM string
            reg_name (str): name of register
            reg_size (int): original size of register
            unused_indices (set): set of unused indices

        Returns:
            str: updated qasm string"""
        required_size = reg_size - len(unused_indices)

        new_id = 0
        qubit_map = {}
        for idx in range(reg_size):
            if idx not in unused_indices:
                # idx -> new_id
                qubit_map[idx] = new_id
                new_id += 1

        # old_id WILL NEVER match the declaration
        # as it will be < the original size of register

        # 1. Replace the qubits first
        #    as the regex may match the new declaration itself
        for old_id, new_id in qubit_map.items():
            if old_id != new_id:
                qasm_str = re.sub(rf"{reg_name}\s*\[{old_id}\]", f"{reg_name}[{new_id}]", qasm_str)

        # 2. Replace the declaration
        qasm_str = re.sub(
            rf"qreg\s+{reg_name}\s*\[{reg_size}\]\s*;",
            f"qreg {reg_name}[{required_size}];",
            qasm_str,
        )
        qasm_str = re.sub(
            rf"qubit\s*\[{reg_size}\]\s*{reg_name}\s*;",
            f"qubit[{required_size}] {reg_name};",
            qasm_str,
        )
        # 1 qubit register can never be partially used :)

        return qasm_str

    @auto_reparse
    def populate_idle_qubits(self) -> None:
        """Converts OpenQASM 3 string to contiguous qasm3 string with gate expansion.

        No loops OR custom functions supported at the moment.
        """
        # Analyse the qasm3 string for registers and find unused qubits
        qubit_indices = self._get_unused_qubit_indices()
        expansion_qasm = ""

        # Add an identity gate for the unused qubits
        for reg, indices in qubit_indices.items():
            for index in indices:
                expansion_qasm += f"i {reg}[{index}];\n"

        qasm: str = self.program
        self._program = qasm + expansion_qasm

    @auto_reparse
    def remove_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, reduces dimension accordingly."""
        qasm_str: str = self.program
        qreg_list = set(self.qubits)
        qubit_indices = self._get_unused_qubit_indices()
        for reg, indices in qubit_indices.items():
            size = 1
            for qreg in qreg_list:
                if qreg[0] == reg:
                    size = qreg[1] or 1
                    break

            # remove the register declarations which are not used
            if len(indices) == size:
                qasm_str = re.sub(rf"qreg\s+{reg}\s*\[\d+\]\s*;", "", qasm_str)
                qasm_str = re.sub(rf"qubit\s*\[\d+\]\s*{reg}\s*;", "", qasm_str)
                if size == 1:
                    qasm_str = re.sub(rf"qubit\s+{reg}\s*;", "", qasm_str)
                try:
                    qreg_list.remove((reg, size))
                except KeyError:
                    qreg_list.remove((reg, None))

            # resize and re-map the indices of the partially used register
            elif len(indices):
                qasm_str = self._remap_qubits(qasm_str, reg, size, indices)
        self._program = qasm_str

    def _validate_qubit_mapping(self, qubit_decls, qubit_mapping: dict):
        """Validate the supplied qubit map
            qubit mapping structure should be like -
                {
                <reg name> : { old_id : new_id, old_id : new_id, ... },
                ...
                }
        Moreover, every reg should be present in the mapping, even if not being remapped.
        The mapping should be complete and indices should be unique and in range.


        Args:
            qubit_decls (list): Qubit register declarations
            qubit_mapping (dict): A dict containing the qubit mapping for
                                  qasm string
        """

        for name, size in qubit_decls:
            size = 1 if size is None else size

            if name not in qubit_mapping:
                raise ValueError(f"Register {name} not present in the qubit mapping.")

            if not isinstance(qubit_mapping[name], dict):
                raise ValueError(f"Mapping for register {name} is not a dictionary.")

            if len(qubit_mapping[name]) != size:
                raise ValueError(
                    f"Mapping for register {name} is not exact. Map is {qubit_mapping[name]}."
                )

            old_indices = set(range(size))
            new_indices = []

            for idx in old_indices:
                if idx not in qubit_mapping[name]:
                    raise ValueError(f"Index {idx} of register {name} not present in the mapping.")
                if qubit_mapping[name][idx] >= size or qubit_mapping[name][idx] < 0:
                    raise ValueError(
                        f"New index {qubit_mapping[name][idx]} of register {name} is out of range."
                    )
                new_indices.append(qubit_mapping[name][idx])

            if set(new_indices) != old_indices:
                raise ValueError(
                    f"Index map of register {name} is not unique. Map is {qubit_mapping[name]}."
                )

    @auto_reparse
    def apply_qubit_mapping(self, qubit_mapping: dict) -> None:
        """Apply qubit mapping for the qasm program

        Args:
            qubit_mapping (dict): A dict containing the qubit mapping for
                                  qasm string

        Returns:
            str: updated qasm string
        """
        if not qubit_mapping:
            return

        qasm = self._program
        qubit_decls = self.qubits
        self._validate_qubit_mapping(qubit_decls, qubit_mapping)

        # need some placeholder to avoid replacing the same qubit multiple times
        # in case of a CYCLIC mapping

        # Eg. { q : {0:1, 1:0} }
        # In this case if we have
        # cnot q[0], q[1]; and apply the mapping
        # first q[0] -> q[1] and state is -
        # cnot q[1], q[1];
        # second, q[1] -> q[0] and state is -
        # cnot q[0], q[0];

        # this is inconsistent

        marker = "-"
        for name, _ in qubit_decls:
            for old_id, new_id in qubit_mapping[name].items():
                if old_id != new_id:
                    qasm = re.sub(rf"{name}\s*\[{old_id}\]", f"{name}[{marker}{new_id}]", qasm)

        # remove the '-' markers
        for name, _ in qubit_decls:
            qasm = re.sub(rf"{name}\[{marker}", f"{name}[", qasm)

        self._program = qasm

    @auto_reparse
    def replace_reset_with_ops(self) -> None:
        """This function finds all the reset operations in QASM string,
        and replaces them with measurement and conditional X gate operations.

        TODO: Does not account for bits named with identifiers or than 'c'
        """
        qasm_string: str = self.program
        lines = qasm_string.split("\n")
        transformed_lines = []
        classical_bit_counter = 0

        for line in lines:
            if line.startswith("reset"):
                # Extract the qubit name(s) being reset
                qubit_name = line.split(" ")[1].strip(";")

                # Check if the reset is for multiple qubits
                if "[" in qubit_name and "]" in qubit_name:
                    # For array-type qubits, handle them individually
                    base_name = qubit_name.split("[")[0]
                    indices = qubit_name[qubit_name.find("[") + 1 : qubit_name.find("]")].split(",")
                    for index in indices:
                        # Create new measurement operation
                        transformed_lines.append(
                            f"measure {base_name}[{index}] -> c{classical_bit_counter};"
                        )
                        # Create new conditional operation
                        transformed_lines.append(
                            f"if (c{classical_bit_counter} == 1) x {base_name}[{index}];"
                        )
                        # Increment the classical bit counter
                        classical_bit_counter += 1
                else:
                    # For single qubits, just replace directly
                    transformed_lines.append(f"measure {qubit_name} -> c{classical_bit_counter};")
                    transformed_lines.append(f"if (c{classical_bit_counter} == 1) x {qubit_name};")
                    classical_bit_counter += 1
            else:
                transformed_lines.append(line)

        transformed_qasm_string = "\n".join(transformed_lines)

        self._program = transformed_qasm_string

    def reverse_qubit_order(self) -> None:
        """Reverse the order of the qubits in the circuit."""

        qubit_decls = self.qubits

        qubit_mapping = {}
        for reg, size in qubit_decls:
            size = 1 if size is None else size
            qubit_mapping[reg] = {old_id: size - old_id - 1 for old_id in range(size)}

        self.apply_qubit_mapping(qubit_mapping)

    @auto_reparse
    def transform(self, device: qbraid.runtime.QuantumDevice, **kwargs) -> None:
        """Transform program to according to device target profile."""
        basis_gates = device.profile.get("basis_gates")

        if basis_gates is not None and len(basis_gates) > 0:
            transformed_qasm = rebase(self.program, basis_gates, **kwargs)
            self._program = normalize_qasm_gate_params(transformed_qasm)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining abstract QuantumJob Class

"""
from __future__ import annotations

from abc import ABC, abstractmethod
from time import sleep, time
from typing import TYPE_CHECKING, Any, Optional

from .enums import JobStatus
from .exceptions import JobStateError, ResourceNotFoundError

if TYPE_CHECKING:
    import qbraid.runtime
    from qbraid.runtime.result_data import ResultDataType


class QuantumJob(ABC):
    """Abstract interface for job-like classes."""

    def __init__(
        self, job_id: str, device: Optional[qbraid.runtime.QuantumDevice] = None, **kwargs
    ):
        self._job_id = job_id
        self._device = device
        self._cache_metadata = {"job_id": job_id, **kwargs}

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """Return a unique id identifying the job."""
        return self._job_id

    @property
    def device(self) -> qbraid.runtime.QuantumDevice:
        """Returns the qbraid QuantumDevice object associated with this job."""
        if self._device is None:
            raise ResourceNotFoundError("Job does not have an associated device.")
        return self._device

    def is_terminal_state(self) -> bool:
        """Returns True if job is in final state. False otherwise."""
        terminal_states = JobStatus.terminal_states()
        if self._cache_metadata.get("status", None) in terminal_states:
            return True

        status = self.status()
        return status in terminal_states

    @abstractmethod
    def status(self) -> JobStatus:
        """Return the status of the job / task , among the values of ``JobStatus``."""

    def metadata(self) -> dict[str, Any]:
        """Return the metadata regarding the job."""
        status = self.status()
        self._cache_metadata["status"] = status
        return self._cache_metadata

    def wait_for_final_state(self, timeout: Optional[int] = None, poll_interval: int = 5) -> None:
        """Poll the job status until it progresses to a final state.

        Args:
            timeout: Seconds to wait for the job. If ``None``, wait indefinitely.
            poll_interval: Seconds between queries. Defaults to 5 seconds.

        Raises:
            JobStateError: If the job does not reach a final state before the specified timeout.

        """
        start_time = time()
        while not self.is_terminal_state():
            elapsed_time = time() - start_time
            if timeout is not None and elapsed_time >= timeout:
                raise JobStateError(f"Timeout while waiting for job {self.id}.")
            sleep(poll_interval)

    @abstractmethod
    def result(self) -> qbraid.runtime.Result[ResultDataType]:
        """Return the results of the job."""

    @abstractmethod
    def cancel(self) -> None:
        """Attempt to cancel the job."""

    def __repr__(self) -> str:
        """String representation of a QuantumJob object."""
        return f"<{self.__class__.__name__}(id:'{self.id}')>"
